"""schema module."""

from bfb_delivery.lib.schema.schema import (  # noqa: F403
    CircuitPlansFromDict,
    CircuitPlansOut,
    CircuitPlansTransformIn,
    CircuitRoutesTransformInFromDict,
    CircuitRoutesTransformOut,
    CircuitRoutesWriteIn,
    CircuitRoutesWriteInAllHHs,
    CircuitRoutesWriteOut,
    DriversAssignDriverIn,
    DriversAssignDriversIn,
    DriversGetAllDriversOut,
    PlansAssignDriverIn,
    PlansAssignDriversIn,
    PlansAssignDriversOut,
    PlansAssignDriversToPlansOut,
    PlansBuildStopsIn,
    PlansConfirmOptimizationsIn,
    PlansConfirmOptimizationsOut,
    PlansCreatePlansOut,
    PlansDistributeRoutesIn,
    PlansDistributeRoutesOut,
    PlansInitializePlansIn,
    PlansInitializePlansOut,
    PlansOptimizeRoutesIn,
    PlansOptimizeRoutesOut,
    PlansUploadSplitChunkedOut,
    PlansUploadStopsIn,
    PlansUploadStopsOut,
    Stops,
)



"""Utility functions for the dispatch module."""

import logging
import os
from os import getcwd as os_getcwd  # For test patching.
from pathlib import Path

from dotenv import load_dotenv
from typeguard import typechecked

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@typechecked
def get_circuit_key() -> str:
    """Get the Circuit API key."""
    load_dotenv(dotenv_path=Path(os_getcwd()) / ".env", override=True)

    key = os.getenv("CIRCUIT_API_KEY")
    if not key:
        raise ValueError(
            "Circuit API key not found. Set the CIRCUIT_API_KEY environment variable."
        )

    return key



"""Public functions wrap internal functions which wrap library functions.

This allows separation of API from implementation. It also allows a simplified public API
separate from a more complex internal API with more options for power users.
"""

from pathlib import Path

from typeguard import typechecked

from bfb_delivery.api import internal
from bfb_delivery.lib.constants import DocStrings


def build_routes_from_chunked(  # noqa: D103
    input_path: str,
    output_dir: str = DocStrings.BUILD_ROUTES_FROM_CHUNKED.defaults["output_dir"],
    start_date: str = DocStrings.BUILD_ROUTES_FROM_CHUNKED.defaults["start_date"],
    no_distribute: bool = DocStrings.BUILD_ROUTES_FROM_CHUNKED.defaults["no_distribute"],
    verbose: bool = DocStrings.BUILD_ROUTES_FROM_CHUNKED.defaults["verbose"],
    extra_notes_file: str = DocStrings.BUILD_ROUTES_FROM_CHUNKED.defaults["extra_notes_file"],
) -> Path:
    return internal.build_routes_from_chunked(
        input_path=input_path,
        output_dir=output_dir,
        start_date=start_date,
        no_distribute=no_distribute,
        verbose=verbose,
        extra_notes_file=extra_notes_file,
    )


build_routes_from_chunked.__doc__ = DocStrings.BUILD_ROUTES_FROM_CHUNKED.api_docstring


@typechecked
def split_chunked_route(  # noqa: D103
    input_path: Path | str,
    output_dir: Path | str = DocStrings.SPLIT_CHUNKED_ROUTE.defaults["output_dir"],
    output_filename: str = DocStrings.SPLIT_CHUNKED_ROUTE.defaults["output_filename"],
    n_books: int = DocStrings.SPLIT_CHUNKED_ROUTE.defaults["n_books"],
    book_one_drivers_file: str = DocStrings.SPLIT_CHUNKED_ROUTE.defaults[
        "book_one_drivers_file"
    ],
    date: str = DocStrings.SPLIT_CHUNKED_ROUTE.defaults["date"],
) -> list[Path]:
    return internal.split_chunked_route(
        input_path=input_path,
        output_dir=output_dir,
        output_filename=output_filename,
        n_books=n_books,
        book_one_drivers_file=book_one_drivers_file,
        date=date,
    )


split_chunked_route.__doc__ = DocStrings.SPLIT_CHUNKED_ROUTE.api_docstring


@typechecked
def create_manifests_from_circuit(  # noqa: D103
    start_date: str = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults["start_date"],
    end_date: str = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults["end_date"],
    plan_ids: list[str] = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults["plan_ids"],
    output_dir: str = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults["output_dir"],
    output_filename: str = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults[
        "output_filename"
    ],
    circuit_output_dir: str = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults[
        "circuit_output_dir"
    ],
    all_hhs: bool = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults["all_hhs"],
    verbose: bool = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults["verbose"],
    extra_notes_file: str = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.defaults[
        "extra_notes_file"
    ],
) -> tuple[Path, Path]:
    final_manifest_path, new_circuit_output_dir = internal.create_manifests_from_circuit(
        start_date=start_date,
        end_date=end_date,
        plan_ids=plan_ids,
        output_dir=output_dir,
        output_filename=output_filename,
        circuit_output_dir=circuit_output_dir,
        all_hhs=all_hhs,
        verbose=verbose,
        extra_notes_file=extra_notes_file,
    )

    return final_manifest_path, new_circuit_output_dir


create_manifests_from_circuit.__doc__ = DocStrings.CREATE_MANIFESTS_FROM_CIRCUIT.api_docstring


@typechecked
def create_manifests(  # noqa: D103
    input_dir: Path | str,
    output_dir: Path | str = DocStrings.CREATE_MANIFESTS.defaults["output_dir"],
    output_filename: str = DocStrings.CREATE_MANIFESTS.defaults["output_filename"],
    extra_notes_file: str = DocStrings.CREATE_MANIFESTS.defaults["extra_notes_file"],
) -> Path:
    formatted_manifest_path = internal.create_manifests(
        input_dir=input_dir,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )

    return formatted_manifest_path


create_manifests.__doc__ = DocStrings.CREATE_MANIFESTS.api_docstring


@typechecked
def combine_route_tables(  # noqa: D103
    input_dir: Path | str,
    output_dir: Path | str = DocStrings.COMBINE_ROUTE_TABLES.defaults["output_dir"],
    output_filename: str = DocStrings.COMBINE_ROUTE_TABLES.defaults["output_filename"],
) -> Path:
    return internal.combine_route_tables(
        input_dir=input_dir, output_dir=output_dir, output_filename=output_filename
    )


combine_route_tables.__doc__ = DocStrings.COMBINE_ROUTE_TABLES.api_docstring


@typechecked
def format_combined_routes(  # noqa: D103
    input_path: Path | str,
    output_dir: Path | str = DocStrings.FORMAT_COMBINED_ROUTES.defaults["output_dir"],
    output_filename: str = DocStrings.FORMAT_COMBINED_ROUTES.defaults["output_filename"],
    extra_notes_file: str = DocStrings.FORMAT_COMBINED_ROUTES.defaults["extra_notes_file"],
) -> Path:
    return internal.format_combined_routes(
        input_path=input_path,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )


format_combined_routes.__doc__ = DocStrings.FORMAT_COMBINED_ROUTES.api_docstring



"""Utility functions for the formatting module."""

import configparser
import logging
import math
import os
from pathlib import Path

import pandas as pd
from openpyxl.cell.cell import Cell
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.worksheet import Worksheet
from typeguard import typechecked

from bfb_delivery.lib.constants import LINE_HEIGHT, BookOneDrivers, Columns, ExtraNotes

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@typechecked
def get_book_one_drivers(file_path: str) -> list[str]:
    """Get the drivers from the book-one driver's file, or the constant if no path.

    Args:
        file_path: Path to the book-one driver's file. If empty, uses a constant list.

    Returns:
        The drivers to include in book one of split chunked routes.
    """
    sheet_one_drivers = [d.value for d in BookOneDrivers]
    if file_path:
        sheet_one_drivers = pd.read_csv(file_path)[Columns.DRIVER].astype(dtype=str).tolist()

    return sheet_one_drivers


@typechecked
def get_extra_notes(file_path: str) -> pd.DataFrame:
    """Get the extra notes from the file, or the constant if no path.

    Args:
        file_path: Path to the extra notes file. If empty, uses a constant DataFrame.

    Returns:
        The extra notes to include in the combined routes.
    """
    extra_notes_df: pd.DataFrame
    if file_path:
        extra_notes_df = pd.read_csv(file_path)
    else:
        extra_notes = ExtraNotes()
        extra_notes_df = extra_notes.df

    validation_sr = extra_notes_df["tag"]
    validation_sr = validation_sr.apply(lambda x: str(x).replace("*", "").strip())
    duplicated_tags = validation_sr[validation_sr.duplicated()].to_list()
    if duplicated_tags:
        raise ValueError(f"Extra notes has duplicated tags: {duplicated_tags}")

    return extra_notes_df


@typechecked
def get_phone_number(key: str, config_path: str = "config.ini") -> str:
    """Get the phone number from the config file.

    Args:
        key: The key in the config file.
        config_path: The path to the config file.

    Returns:
        The phone number.
    """
    section_key = "phone_numbers"
    phone_number = (
        "NO PHONE NUMBER. "
        "See warning in logs for instructions on setting up your config file."
    )
    full_config_path = Path(config_path)
    full_config_path = full_config_path.resolve()
    config_instructions = (
        f"In config file, under '[{section_key}]', add '{key} = (555) 555-5555'."
    )

    if os.path.exists(full_config_path):
        config = configparser.ConfigParser()
        try:
            config.read(full_config_path)
            phone_number = config[section_key][key]
        except KeyError:
            logger.warning(
                f"{key} not found in config file: {full_config_path}. {config_instructions}"
            )
    else:
        logger.warning(
            (
                f"Config file not found: {full_config_path}. "
                f"Create the file. {config_instructions}"
            )
        )

    return str(phone_number)


@typechecked
def map_columns(df: pd.DataFrame, column_name_map: dict[str, str], invert_map: bool) -> None:
    """Map column names in a DataFrame.

    Operates in place.

    Args:
        df: The DataFrame to map.
        column_name_map: The mapping of column names.
        invert_map: Whether to invert the mapping.
    """
    if invert_map:
        column_name_map = {v: k for k, v in column_name_map.items()}

    df.rename(columns=column_name_map, inplace=True)


@typechecked
def set_row_height_of_wrapped_cell(cell: Cell) -> None:
    """Set appropriate row height for a cell with wrapped text.

    Inspects cell properties (value, font, column width) and calculates the
    appropriate row height for wrapped text display. Works with both merged
    and non-merged cells.

    Args:
        cell: The cell object containing wrapped text.

    Note:
        This calculation is approximate. The actual row height needed depends on
        font, size, and formatting. Manual review may be needed for precise results.
        Modifies the row height in place.
    """
    if cell.value is not None:
        ws: Worksheet = cell.parent
        row_num: int = cell.row

        char_width: float = (
            1.2 if (cell.font is not None and getattr(cell.font, "bold", False)) else 1.0
        )

        cell_width: float = _get_cell_width(cell=cell)

        text_length: float = len(str(cell.value)) * char_width
        lines: float = text_length / cell_width
        height: float = max(LINE_HEIGHT, math.ceil(lines) * LINE_HEIGHT)

        ws.row_dimensions[row_num].height = height

    return


@typechecked
def _get_cell_width(cell: Cell) -> float:
    """Get the total width available for a cell, accounting for merges.

    Args:
        cell: The cell to measure.

    Returns:
        The total width available (single column or sum of merged columns).
    """
    ws: Worksheet = cell.parent
    merged_range = next(
        (merged for merged in ws.merged_cells.ranges if cell.coordinate in merged), None
    )

    if merged_range is not None:
        total_width = sum(
            ws.column_dimensions[get_column_letter(col)].width
            for col in range(merged_range.min_col, merged_range.max_col + 1)
        )
    else:
        total_width = ws.column_dimensions[cell.column_letter].width

    return total_width



"""Classes for making API calls."""

import logging

from typeguard import typechecked

from comb_utils import (
    BaseCaller,
    BaseDeleteCaller,
    BaseGetCaller,
    BasePagedResponseGetter,
    BasePostCaller,
)

from bfb_delivery.lib.constants import CIRCUIT_URL, CircuitColumns, RateLimits
from bfb_delivery.lib.dispatch.utils import get_circuit_key

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


# TODO: https://github.com/crickets-and-comb/bfb_delivery/issues/138:
# Why are we using _set_url instead of the url property?
# Why are we using _set_request_call instead of the _request_call property?


class BaseKeyRetriever:
    """A base class for getting the API key.

    Presets the API key to be used for authentication.
    """

    @typechecked
    def _get_API_key(self) -> str:
        """Get the API key.

        Returns:
            The API key.
        """
        return get_circuit_key()


class BaseBFBGetCaller(BaseKeyRetriever, BaseGetCaller):
    """A base class for making GET API calls with BFB Circuit key."""


class BaseBFBPostCaller(BaseKeyRetriever, BasePostCaller):
    """A base class for making POST API calls with BFB Circuit key."""


class BaseBFBDeleteCaller(BaseKeyRetriever, BaseDeleteCaller):
    """A base class for making DELETE API calls with BFB Circuit key."""


class BaseOptimizationCaller(BaseKeyRetriever, BaseCaller):
    """A base class for checking the status of an optimization."""

    #: The ID of the operation.
    operation_id: str
    #: Whether the optimization is finished.
    finished: bool

    _timeout: float = RateLimits.WRITE_TIMEOUT_SECONDS
    _min_wait_seconds: float = RateLimits.OPTIMIZATION_PER_SECOND
    _wait_seconds: float = _min_wait_seconds

    #: The ID of the plan.
    _plan_id: str
    #: The title of the plan.
    _plan_title: str

    @typechecked
    def __init__(self, plan_id: str, plan_title: str) -> None:  # noqa: ANN401
        """Initialize the BaseOptimizationCaller object.

        Args:
            plan_id: The ID of the plan. (e.g. plans/asfoghaev)
            plan_title: The title of the plan.
        """
        self._plan_id = plan_id
        self._plan_title = plan_title
        super().__init__()

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Sets `operation_id` and whether the optimization is `finished`.

        Raises:
            RuntimeError: If the optimization was canceled, stops were skipped, or there were
                errors.
        """
        super()._handle_200()

        if self.response_json[CircuitColumns.METADATA][CircuitColumns.CANCELED]:
            raise RuntimeError(
                f"Optimization canceled for {self._plan_title} ({self._plan_id}):"
                f"\n{self.response_json}"
            )
        if self.response_json.get(CircuitColumns.RESULT):
            if self.response_json[CircuitColumns.RESULT].get(CircuitColumns.SKIPPED_STOPS):
                raise RuntimeError(
                    f"Skipped optimization stops for {self._plan_title} ({self._plan_id}):"
                    f"\n{self.response_json}"
                )
            if self.response_json[CircuitColumns.RESULT].get(CircuitColumns.CODE):
                raise RuntimeError(
                    f"Errors in optimization for {self._plan_title} ({self._plan_id}):"
                    f"\n{self.response_json}"
                )

        self.operation_id = self.response_json[CircuitColumns.ID]
        self.finished = self.response_json[CircuitColumns.DONE]


class PagedResponseGetterBFB(BaseKeyRetriever, BasePagedResponseGetter):
    """Class for getting paged responses."""


class PlanInitializer(BaseBFBPostCaller):
    """Class for initializing plans."""

    #: The ID of the plan.
    plan_id: str
    #: Whether the plan is writeable.
    writable: bool

    #: The data dictionary for the plan.
    _plan_data: dict

    @typechecked
    def __init__(self, plan_data: dict) -> None:
        """Initialize the PlanInitializer object.

        Args:
            plan_data: The data dictionary for the plan.
                To pass to `requests.post` `json` param.
        """
        self._plan_data = plan_data
        self._call_kwargs = {"json": plan_data}
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call."""
        self._url = f"{CIRCUIT_URL}/plans"

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Sets `plan_id` and `writable`.
        """
        super()._handle_200()
        self.plan_id = self.response_json[CircuitColumns.ID]
        self.writable = self.response_json[CircuitColumns.WRITABLE]


class StopUploader(BaseBFBPostCaller):
    """Class for batch uploading stops."""

    stop_ids: list[str]

    _min_wait_seconds: float = RateLimits.BATCH_STOP_IMPORT_SECONDS
    _wait_seconds: float = _min_wait_seconds

    _plan_id: str
    _plan_title: str

    @typechecked
    def __init__(
        self,
        plan_id: str,
        plan_title: str,
        stop_array: list[dict[str, dict[str, str] | list[str] | int | str]],
    ) -> None:
        """Initialize the StopUploader object.

        Args:
            plan_id: The ID of the plan. (e.g. plans/asfoghaev)
            plan_title: The title of the plan.
            stop_array: The array of stops dictionaries to upload.
                To pass to `requests.post` `json` param.
        """
        self._plan_id = plan_id
        self._plan_title = plan_title
        self._stop_array = stop_array
        self._call_kwargs = {"json": stop_array}
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}/stops:import"

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Sets `stop_ids` to the successful stop IDs.

        Raises:
            RuntimeError: If stops failed to upload.
            RuntimeError: If the number of stops uploaded differs from input.
        """
        super()._handle_200()

        self.stop_ids = self.response_json["success"]
        failed = self.response_json.get("failed")
        if failed:
            raise RuntimeError(
                f"For {self._plan_title} ({self._plan_id}), failed to upload stops:\n{failed}"
            )
        elif len(self.stop_ids) != len(self._stop_array):
            raise RuntimeError(
                f"For {self._plan_title} ({self._plan_id}), did not upload same number of "
                f"stops as input:\n{self.stop_ids}\n{self._stop_array}"
            )


class OptimizationLauncher(BaseOptimizationCaller, BaseBFBPostCaller):
    """A class for launching route optimization.

    Args:
        plan_id: The ID of the plan. (e.g. plans/asfoghaev)
        plan_title: The title of the plan.
    """

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with the `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}:optimize"


class OptimizationChecker(BaseOptimizationCaller, BaseBFBGetCaller):
    """A class for checking the status of an optimization."""

    _timeout: float = RateLimits.READ_TIMEOUT_SECONDS
    _min_wait_seconds: float = RateLimits.READ_SECONDS
    _wait_seconds: float = _min_wait_seconds

    @typechecked
    def __init__(self, plan_id: str, plan_title: str, operation_id: str) -> None:
        """Initialize the OptimizationChecker object.

        Args:
            plan_id: The ID of the plan.
            plan_title: The title of the plan.
            operation_id: The ID of the operation.
        """
        self.operation_id = operation_id
        super().__init__(plan_id=plan_id, plan_title=plan_title)

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call."""
        self._url = f"{CIRCUIT_URL}/{self.operation_id}"


class PlanDistributor(BaseBFBPostCaller):
    """Class for distributing plans."""

    distributed: bool

    #: The ID of the plan.
    _plan_id: str
    #: The title of the plan
    _plan_title: str

    @typechecked
    def __init__(self, plan_id: str, plan_title: str) -> None:
        """Initialize the PlanDistributor object.

        Args:
            plan_id: The ID of the plan. (e.g. plans/asfoghaev)
            plan_title: The title of the plan.
        """
        self._plan_id = plan_id
        self._plan_title = plan_title
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with the `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}:distribute"

    @typechecked
    def _handle_200(self) -> None:
        """Handle a 200 response.

        Raises:
            RuntimeError: If the plan was not distributed.
        """
        super()._handle_200()
        self.distributed = self.response_json[CircuitColumns.DISTRIBUTED]
        if not self.distributed:
            raise RuntimeError(
                f"Failed to distribute plan {self._plan_title} ({self._plan_id}):"
                f"\n{self.response_json}"
            )


class PlanDeleter(BaseBFBDeleteCaller):
    """Class for deleting plans."""

    #: Whether the plan was deleted.
    deletion: bool = False

    @typechecked
    def __init__(self, plan_id: str) -> None:
        """Initialize the PlanDeleter object.

        Args:
            plan_id: The ID of the plan.
        """
        self._plan_id = plan_id
        super().__init__()

    @typechecked
    def _set_url(self) -> None:
        """Set the URL for the API call with the `plan_id`."""
        self._url = f"{CIRCUIT_URL}/{self._plan_id}"

    @typechecked
    def _handle_204(self) -> None:
        """Handle a 204 response.

        Sets `deletion` to True.
        """
        super()._handle_204()
        self.deletion = True



# noqa: D100
__doc__ = """
.. click:: bfb_delivery.cli.format_combined_routes:main
   :prog: format_combined_routes
   :nested: full
"""
import logging

import click
from typeguard import typechecked

from bfb_delivery import format_combined_routes
from bfb_delivery.lib.constants import DocStrings

logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(levelname)s - %(message)s")
logger = logging.getLogger(__name__)


@click.command(help=DocStrings.FORMAT_COMBINED_ROUTES.cli_docstring)
@click.option("--input_path", required=True, help="The path to the combined routes table.")
@click.option(
    "--output_dir",
    type=str,
    required=False,
    default=DocStrings.FORMAT_COMBINED_ROUTES.defaults["output_dir"],
    help=DocStrings.FORMAT_COMBINED_ROUTES.args["output_dir"],
)
@click.option(
    "--output_filename",
    type=str,
    required=False,
    default=DocStrings.FORMAT_COMBINED_ROUTES.defaults["output_filename"],
    help=DocStrings.FORMAT_COMBINED_ROUTES.args["output_filename"],
)
@click.option(
    "--extra_notes_file",
    type=str,
    required=False,
    default=DocStrings.FORMAT_COMBINED_ROUTES.defaults["extra_notes_file"],
    help=DocStrings.FORMAT_COMBINED_ROUTES.args["extra_notes_file"],
)
@typechecked
def main(  # noqa: D103
    input_path: str, output_dir: str, output_filename: str, extra_notes_file: str
) -> str:
    path = format_combined_routes(
        input_path=input_path,
        output_dir=output_dir,
        output_filename=output_filename,
        extra_notes_file=extra_notes_file,
    )
    logger.info(f"Formatted driver manifest saved to:\n{path.resolve()}")

    return str(path)
